/* cat.c - concatenate files and print on the standard output

   AUTHOR: Gregory Pietsch
   
*/

#include <ctype.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include "getopt.h"
#include "config.h"

/* flags */

#define F_NUMBER_NONBLANK       (1)
#define F_NUMBER                (2)
#define F_SQUEEZE_BLANK         (4)
#define F_SHOW_NONPRINTING      (8)
#define F_SHOW_ENDS             (0x10)
#define F_SHOW_TABS             (0x20)

/* globals */
unsigned flags = 0U;
int show_help = 0, show_version = 0;
static char *shortopts = "benstuvAET";
static struct option longopts[] = {
  {"number-nonblank", 0, 0, 'b'},
  {"number", 0, 0, 'n'},
  {"squeeze-blank", 0, 0, 's'},
  {"show-nonprinting", 0, 0, 'v'},
  {"show-all", 0, 0, 'A'},
  {"show-ends", 0, 0, 'E'},
  {"show-tabs", 0, 0, 'T'},
  {"help", 0, &show_help, 1},
  {"version", 0, &show_version, 1},
  {0, 0, 0, 0}
};

/* functions */

#ifndef HAVE_ISASCII
/* No isascii, so I have to make it explicit. */
static int
isascii (int c)
{
  return (c >= 0 && c <= 127);
}
#endif

/* help: show the help message */
void
help (void)
{
#include "help.h"
}

/* version: show the version message */
void
version (void)
{
  puts (PACKAGE_STRING);
}

/* parse_args: parse the command-line arguments */
void
parse_args (int c, char *v[])
{
  int o;
  while ((o = getopt_long (c, v, shortopts, longopts, 0)) != EOF)
    {
      switch (o)
	{
	case 'b':		/* --number-nonblank */
	  flags |= F_NUMBER_NONBLANK;
	  break;
	case 'e':
	  flags |= F_SHOW_NONPRINTING | F_SHOW_ENDS;
	  break;
	case 'n':		/* --number */
	  flags |= F_NUMBER;
	  break;
	case 's':		/* --squeeze-blank */
	  flags |= F_SQUEEZE_BLANK;
	  break;
	case 't':
	  flags |= F_SHOW_NONPRINTING | F_SHOW_TABS;
	  break;
	case 'u':
	  break;
	case 'v':		/* --show-nonprinting */
	  flags |= F_SHOW_NONPRINTING;
	  break;
	case 'A':		/* --show-all */
	  flags |= F_SHOW_NONPRINTING | F_SHOW_ENDS | F_SHOW_TABS;
	  break;
	case 'E':		/* --show-ends */
	  flags |= F_SHOW_ENDS;
	  break;
	case 'T':		/* --show-tabs */
	  flags |= F_SHOW_TABS;
	  break;
	case 0:		/* --help and --version */
	  if (show_help)
	    {
	      help ();
	      exit (EXIT_FAILURE);
	    }
	  if (show_version)
	    version ();
	  exit (EXIT_SUCCESS);
	default:
	  break;
	}
    }
  if ((flags & (F_NUMBER_NONBLANK | F_NUMBER)) ==
      (F_NUMBER_NONBLANK | F_NUMBER))
    flags &= ~F_NUMBER;		/* --number-nonblank takes precedence */
}

/* output a character */
void
output_char (int c)
{
  if (c == '\n')
    {
      if (flags & F_SHOW_ENDS)
	putchar ('$');
      putchar (c);
    }
  else if (c == '\t' && (flags & F_SHOW_TABS))
    fputs ("^I", stdout);
  else if (c != '\t' && (flags & F_SHOW_NONPRINTING))
    {
      if (!isascii (c))
	{
	  fputs ("M-", stdout);
	  c &= UCHAR_MAX >> 1;
	}
      if (isprint (c))
	putchar (c);
      else
	{
	  putchar ('^');
	  putchar (c ^ 64);     /* ASCII hack */
	}
    }
  else
    putchar (c);
}

/* concatenate one file - return 0 on success */
int
cat (char *name)
{
  FILE *f = 0;
  int f_stdin = 0;
  int f_binary = 0;
  int f_prev_blank = 0;
  unsigned long lineno = 0U;
  char s[BUFSIZ], *p = 0;
  int c;

  /* determine input file */
  if (*name == '-' && name[1] == '\0')
    {
      f = stdin;
      f_stdin = 1;
    }
  else if (flags &
	   (F_NUMBER_NONBLANK | F_NUMBER | F_SQUEEZE_BLANK | F_SHOW_ENDS))
    f = fopen (name, "r");
  else
    {
      f = fopen (name, "rb");
      f_binary = 1;
    }
  if (f == 0)
    {
      fprintf (stderr, "cat: Can't open \'%s\'\n", name);
      return 1;
    }
  if (f_binary)
    while ((c = getc (f)) != EOF)
      /* only have to worry about --show-tabs and --show-nonprinting */
      output_char (c);
  else
    /* have to deal with everything */
    while (!feof (f))
      {
	if (fgets (s, BUFSIZ, f))
	  {
	    /* is it a blank line? */
	    if (*s == '\n')
	      {
		if ((flags & F_SQUEEZE_BLANK) & f_prev_blank)
		  ;
		else
		  {
		    if (!(flags & F_NUMBER_NONBLANK))
		      ++lineno;
		    if (flags & F_NUMBER)
		      printf ("%6lu  ", lineno);
		    output_char (*s);
		    f_prev_blank = 1;
		  }
	      }
	    else
	      {
		++lineno;
		if (flags & (F_NUMBER | F_NUMBER_NONBLANK))
		  printf ("%6lu  ", lineno);
		p = s;
		while (p && *p != '\n')
		  {
		    if (*p == '\0')
		      p = fgets (s, BUFSIZ, f);
		    else
		      {
			output_char (*p);
			++p;
		      }
		  }
		output_char ('\n');
		f_prev_blank = 0;
	      }
	  }
      }
  if (!f_stdin)
    fclose (f);
  return 0;
}

int
main (int c, char *v[])
{
  parse_args (c, v);
  if (v[optind])
    {
      while (v[optind])
	cat (v[optind++]);
    }
  else
    cat ("-");
  return 0;
}

/* END OF FILE */
